﻿using SimpleJSON;
using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.Text;
using ToySerialController.UI;
using ToySerialController.Utils;
using UnityEngine;
using VAMLaunchPlugin;
using VAMLaunchPlugin.MotionSources;

namespace ToySerialController.Device.OutputTarget
{
    public class VamLaunchOutputTarget : IOutputTarget
    {
        private UIHorizontalGroup ButtonGroup;

        private VAMLaunchNetwork _network;
        private float _networkPollTimer;

        private float _simulatorTarget;
        private float _simulatorSpeed;

        private const string SERVER_IP = "127.0.0.1";
        private const int SERVER_LISTEN_PORT = 15600;
        private const int SERVER_SEND_PORT = 15601;
        private const float NETWORK_LISTEN_INTERVAL = 0.033f;

        private int _lastWrittenPos = 0;
        private int _lastWrittenDelta = 0;
        private byte _lastSentLaunchPos = 0;
        private float _udpatesSinceLastTracking = 0f;
        private float _slowInitiate = 0f;


        private JSONStorableFloat _simulatorPosition;

        private JSONStorableFloat _minPosition;
        private JSONStorableFloat _maxPosition;
        private JSONStorableFloat _positionSampleRate;
        private JSONStorableFloat _minLaunchSignalTimeThreshold;
        private JSONStorableFloat _maxLaunchSignalTimeThreshold;
        private JSONStorableFloat _lowerVelocityBarrier;
        private JSONStorableFloat _higherVelocityBarrier;
        private JSONStorableFloat _launchSpeedMultiplier;
        private JSONStorableFloat _launchInputMultiplier;

        private JSONStorableFloat _launchMaxSpeed;
        private JSONStorableFloat _launchMinSpeed;

        //private JSONStorableBool _enableDynamicMinMaxRange;
        //private JSONStorableFloat _currentDynamicMaxRangePosition;
        //private JSONStorableFloat _currentDynamicMinRangePosition;

        private JSONStorableFloat _currentLaunchSignalTimeThreshold;
        private JSONStorableFloat _currentTargetPos;

        private JSONStorableStringChooser _curveAxisChooser;
        private JSONStorableBool _invertToggle;

        private JSONStorableFloat _launchRandomMaxRangeA;
        private JSONStorableFloat _launchRandomMaxRangeB;
        private JSONStorableBool _enableRandomMaxRange;

        private ZoneSimulatorOptions _currentZoneSimulatorOptions;
        private ILaunchMotionSource _currentMotionSource;

        private UIDynamic _spacerA, _spacerB, _spacerC, _spacerD;

        private UIGroup _group;
        private UIGroup _launchGroup;
        private bool _launchGroupVisible;

        private bool _lastRandomEnabled = false;
        private float _lastRandomOriginalDeviceMaxRange = 99f;

        private float _lastRndTime = 0f;
        private float _currentRndMaxValue = 0f;

        private bool _tracking = false;
        private bool _resetting = false;
        private int _loadingStatus = LOAD_STATUS_NONE;
        private const int LOAD_STATUS_NONE = 0;
        private const int LOAD_STATUS_SCENE_LOADING = 1;
        private const int LOAD_STATUS_WAIT_ONE_WRITE = 2;
        private const int LOAD_STATUS_WAIT_ONE_UPDATE = 3;

        private bool _bigDeltaDetected = false;


        public VamLaunchOutputTarget()
        {
            // _lagFactor = new FpsLagFactor();
            _currentZoneSimulatorOptions = new ZoneSimulatorOptions();
            _currentMotionSource = new ZoneSimulator(_currentZoneSimulatorOptions);
        }

        public void CreateUI(IUIBuilder builder)
        {   
            ButtonGroup = builder.CreateHorizontalGroup(510, 50, new Vector2(10, 0), 2, idx => builder.CreateButtonEx());
            var startSerialButton = ButtonGroup.items[0].GetComponent<UIDynamicButton>();
            startSerialButton.label = "Start Udp/Launch";
            startSerialButton.button.onClick.AddListener(StartUdp);

            var stopSerialButton = ButtonGroup.items[1].GetComponent<UIDynamicButton>();
            stopSerialButton.label = "Stop Udp/Launch";
            stopSerialButton.button.onClick.AddListener(StopUdp);

            _currentTargetPos = builder.CreateSlider("launchTargetPos", "Current Axis Input", 0.0f, 0.0f, 99.99f, true, false, false);
            _simulatorPosition = builder.CreateSlider("launchSimulator", "Device Simulator", 0.0f, 0.0f, 99.0f, true, false, false);
            _maxPosition = builder.CreateSlider("launchMaxPosition", "Device Max Range", 80.0f, 0.0f, 99.0f, true, true);
            _minPosition = builder.CreateSlider("launchMinPosition", "Device Min Range", 0.0f, 0.0f, 80.0f, true, true);

            _group = new UIGroup(builder);
            _launchGroup = new UIGroup(_group);
            _launchGroupVisible = true;
            _group.CreateButton("Other Launch Options", () => _launchGroup.SetVisible(_launchGroupVisible = !_launchGroupVisible), new Color(0.3f, 0.3f, 0.3f), Color.white);

            _enableRandomMaxRange = _launchGroup.CreateToggle("launchRandomMaxRangeEnabled", "Use Random Max Range", false);
            _launchRandomMaxRangeA = _launchGroup.CreateSlider("launchRandomMaxRangeA", "Random Max Range (Min Rnd Value)", 50.0f, 10.0f, 200.0f, true, true);
            _launchRandomMaxRangeB = _launchGroup.CreateSlider("launchRandomMaxRangeB", "Random Max Range (Max Rnd Value)", 100.0f, 10.0f, 200.0f, true, true);
            
            _spacerA = _launchGroup.CreateSpacer(10, false);

            //_enableDynamicMinMaxRange = _launchGroup.CreateToggle($"launchDynamicRangeEnabled", "Use Device Dynamic Range", false);
            //_currentDynamicMaxRangePosition = _launchGroup.CreateSlider("launchDynamicRangeMax", "Current Device Dynamic Max Range", 80.0f, 10.0f, 99.0f, true, false, false);
            //_currentDynamicMinRangePosition = _launchGroup.CreateSlider("launchDynamicRangeMin", "Current Device Dynamic Min Range", 0.0f, 0.0f, 80.0f, true, false, false);

            _spacerB = _launchGroup.CreateSpacer(10, false);

            _launchSpeedMultiplier = _launchGroup.CreateSlider("launchSpeedMultiplier", "Device Speed Multiplier", 1.0f, 0.01f, 2.0f, true, true);
            _launchInputMultiplier = _launchGroup.CreateSlider("launchInputMultiplier", "Axis Input Multiplier", 1.0f, 0.1f, 2.0f, true, true);
            _spacerC = _launchGroup.CreateSpacer(10, false);

            _currentLaunchSignalTimeThreshold = _launchGroup.CreateSlider("launchAdjustTime", "Current Adjust Time", 0.15f, 0.0f, 0.40f, true, false, false);
            _minLaunchSignalTimeThreshold = _launchGroup.CreateSlider("launchMinSignalTime", "Min Adjust Time Treshold", 0.12f, 0.0f, 0.40f, true, true);
            _maxLaunchSignalTimeThreshold = _launchGroup.CreateSlider("launchMaxSignalTime", "Max Adjust Time Treshold", 0.25f, 0.0f, 0.40f, true, true);
            _lowerVelocityBarrier = _launchGroup.CreateSlider("launchLowerVelBarrier", "Min Adjust Time Vel Barrier", 20.0f, 0.0f, 99.0f, true, true);
            _higherVelocityBarrier = _launchGroup.CreateSlider("launchHigherVelbarrier", "Max Adjust Time Vel Barrier", 50.0f, 0.0f, 99.0f, true, true);

            _spacerD = _launchGroup.CreateSpacer(10, false);
            _positionSampleRate = _launchGroup.CreateSlider("launchPositionSampleRate", "Position Sample Rate", 40.0f, 10.0f, 90.0f, true, true);
            _curveAxisChooser = _launchGroup.CreateScrollablePopup($"launchCurveAxis", "Curve Axis Input", new List<string> { "L0", "L1", "L2", "R0", "R1", "R2", "V0", "A0", "A1", "A2" }, "L0", CurveAxisChooserCallback, false);
            _invertToggle = _launchGroup.CreateToggle($"launchInvertToggle", "Invert Axis Input", false);
            _launchMaxSpeed = _launchGroup.CreateSlider("launchMaxSpeed", "Device Max Speed", Mathf.Min(70f, LaunchUtils.LAUNCH_MAX_SPEED), LaunchUtils.LAUNCH_MIN_SPEED, LaunchUtils.LAUNCH_MAX_SPEED, true, true);
            _launchMinSpeed = _launchGroup.CreateSlider("launchMinSpeed", "Device Min Speed", Mathf.Min(10f, LaunchUtils.LAUNCH_MIN_SPEED), LaunchUtils.LAUNCH_MIN_SPEED, 30f, true, true);

            _launchGroup.SetVisible(_launchGroupVisible);
        }

        protected void CurveAxisChooserCallback(string val)
        {
        }

        private void PrepareOptions()
        {
            // _currentZoneSimulatorOptions.CurrentTargetPos = _currentTargetPos.val;
            _currentZoneSimulatorOptions.CurrentLaunchSignalTimeThreshold = _currentLaunchSignalTimeThreshold.val;

            _currentZoneSimulatorOptions.PositionSampleRate = _positionSampleRate.val;
            _currentZoneSimulatorOptions.MinLaunchSignalTimeThreshold = _minLaunchSignalTimeThreshold.val;
            _currentZoneSimulatorOptions.MaxLaunchSignalTimeThreshold = _maxLaunchSignalTimeThreshold.val;
            _currentZoneSimulatorOptions.LowerVelocityBarrier = _lowerVelocityBarrier.val;
            _currentZoneSimulatorOptions.HigherVelocityBarrier = _higherVelocityBarrier.val;
            _currentZoneSimulatorOptions.LaunchSpeedMultiplier = _launchSpeedMultiplier.val;
            _currentZoneSimulatorOptions.InputMultiplier = _launchInputMultiplier.val;
            _currentZoneSimulatorOptions.MinPosition = _minPosition.val;
            _currentZoneSimulatorOptions.MaxPosition = _maxPosition.val;
            
            if (_enableRandomMaxRange.val)
            {
                if (!_lastRandomEnabled)
                {
                    _lastRandomOriginalDeviceMaxRange = _maxPosition.val;
                }
                if ((Time.time - _lastRndTime) > 1f)
                {
                    if (_currentZoneSimulatorOptions.CurrentSign == -1 && _simulatorTarget < _launchRandomMaxRangeA.val) // _currentZoneSimulatorOptions.CurrentTargetPos < _launchRandomMaxRangeA.val
                    {
                        _lastRndTime = Time.time;
                        float rnd = Mathf.Min(LaunchUtils.LAUNCH_MAX_VAL, _launchRandomMaxRangeA.val + (_launchRandomMaxRangeB.val - _launchRandomMaxRangeA.val) * UnityEngine.Random.value);
                        _currentRndMaxValue = Mathf.Max(rnd, _minPosition.val);
                    }
                    else
                    {
                        // try again in a moment
                        _lastRndTime = Time.time - 0.8f;
                    }
                }
                _maxPosition.val = _currentRndMaxValue;
                _currentZoneSimulatorOptions.MaxPosition = _currentRndMaxValue;
            } 
            else
            {
                _lastRndTime = Time.time;
                _currentRndMaxValue = _maxPosition.val;
                if (_lastRandomEnabled)
                {
                    _maxPosition.val = _lastRandomOriginalDeviceMaxRange;
                }
            }
            _lastRandomEnabled = _enableRandomMaxRange.val;

            // Dynamic Range never turned out any good, disable it
            _currentZoneSimulatorOptions.DynamicRangeMin = 0f; // _currentDynamicMinRangePosition.val;
            _currentZoneSimulatorOptions.DynamicRangeMax = 99f; // _currentDynamicMaxRangePosition.val;
            _currentZoneSimulatorOptions.DynamicRangeEnabled = false; // _enableDynamicMinMaxRange.val;

            // LaunchUtils.LAUNCH_MAX_SPEED_CALC = _launchMaxSpeed.val;
            LaunchUtils.LAUNCH_MIN_SPEED_CALC = _launchMinSpeed.val; // only adjust min-speed which can affect calculations. Max speed already capped in other ways.
        }
        private void SetOptions()
        {
            // writeback results
            _currentTargetPos.val = _currentZoneSimulatorOptions.CurrentTargetPos;
            _currentLaunchSignalTimeThreshold.val = _currentZoneSimulatorOptions.CurrentLaunchSignalTimeThreshold;

            //_currentDynamicMinRangePosition.val = _currentZoneSimulatorOptions.DynamicRangeMin;
            //_currentDynamicMaxRangePosition.val = _currentZoneSimulatorOptions.DynamicRangeMax;
        }

        public void DestroyUI(IUIBuilder builder)
        {
            builder.Destroy(ButtonGroup);
            
            builder.Destroy(_simulatorPosition);

            builder.Destroy(_minPosition);
            builder.Destroy(_maxPosition);
            builder.Destroy(_enableRandomMaxRange);
            builder.Destroy(_launchRandomMaxRangeA);
            builder.Destroy(_launchRandomMaxRangeB);
            builder.Destroy(_positionSampleRate);
            builder.Destroy(_minLaunchSignalTimeThreshold);
            builder.Destroy(_maxLaunchSignalTimeThreshold);
            builder.Destroy(_lowerVelocityBarrier);
            builder.Destroy(_higherVelocityBarrier);
            builder.Destroy(_launchSpeedMultiplier);
            builder.Destroy(_launchInputMultiplier);
            //builder.Destroy(_enableDynamicMinMaxRange);
            //builder.Destroy(_currentDynamicMaxRangePosition);
            //builder.Destroy(_currentDynamicMinRangePosition);
            builder.Destroy(_currentLaunchSignalTimeThreshold);
            builder.Destroy(_currentTargetPos);
            builder.Destroy(_invertToggle);
            builder.Destroy(_curveAxisChooser);
            builder.Destroy(_launchMaxSpeed);
            builder.Destroy(_launchMinSpeed);

            builder.Destroy(_spacerA);
            builder.Destroy(_spacerB);
            builder.Destroy(_spacerC);

            builder.Destroy(_group);
            builder.Destroy(_launchGroup);
        }

        public void RestoreConfig(JSONNode config)
        {
            config.Restore(_minPosition);
            config.Restore(_maxPosition);
            config.Restore(_positionSampleRate);
            config.Restore(_minLaunchSignalTimeThreshold);
            config.Restore(_maxLaunchSignalTimeThreshold);
            config.Restore(_lowerVelocityBarrier);
            config.Restore(_higherVelocityBarrier);
            config.Restore(_launchSpeedMultiplier);
            config.Restore(_launchInputMultiplier);
            config.Restore(_curveAxisChooser);
            config.Restore(_invertToggle);
            //config.Restore(_enableDynamicMinMaxRange);
            config.Restore(_launchMaxSpeed);
            config.Restore(_launchMinSpeed);
            config.Restore(_launchRandomMaxRangeA);
            config.Restore(_launchRandomMaxRangeB);
            config.Restore(_enableRandomMaxRange);
        }

        public void StoreConfig(JSONNode config)
        {
            config.Store(_minPosition);
            config.Store(_maxPosition);
            config.Store(_positionSampleRate);
            config.Store(_minLaunchSignalTimeThreshold);
            config.Store(_maxLaunchSignalTimeThreshold);
            config.Store(_lowerVelocityBarrier);
            config.Store(_higherVelocityBarrier);
            config.Store(_launchSpeedMultiplier);
            config.Store(_launchInputMultiplier);
            config.Store(_curveAxisChooser);
            config.Store(_invertToggle);
            //config.Store(_enableDynamicMinMaxRange);
            config.Store(_launchMaxSpeed);
            config.Store(_launchMinSpeed);
            config.Store(_launchRandomMaxRangeA);
            config.Store(_launchRandomMaxRangeB);
            config.Store(_enableRandomMaxRange);
        }

        private void StartUdp()
        {
            if (_network != null)
                return;

            try
            {
                _network = new VAMLaunchNetwork();
                _network.Init(SERVER_IP, SERVER_LISTEN_PORT, SERVER_SEND_PORT);

                SuperController.LogMessage($"Upd started on port: {SERVER_SEND_PORT}");
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
        }

        private void StopUdp()
        {
            if (_network == null)
                return;

            try
            {
                _network.Stop();
                _network = null;
            }
            catch(Exception e)
            {
                SuperController.LogError(e.ToString());
            }

            SuperController.LogMessage("Upd stopped");
        }

        public void Write(string data)
        {
            // SuperController.LogMessage(data);
            if (_loadingStatus == LOAD_STATUS_SCENE_LOADING)
            {
                return;
            }

            string prt = "";
            foreach(string str in data.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries))
            {
                if (str.Length == 9 && str.StartsWith(_curveAxisChooser.val))
                {
                    prt = str;
                    break;
                }
            }
            if (prt == "")
                return;

            try
            {
                const int BIG_DELTA = 4000;
                int targetPos = int.Parse(prt.Substring(2, 4));

                if (Mathf.Abs(targetPos - _lastWrittenPos) > BIG_DELTA)
                {
                    if (!_bigDeltaDetected)
                    {
                        _bigDeltaDetected = true;
                        //SuperController.LogMessage("BIG_DELTA_DETECTED");
                    }
                    // _lastWrittenPos in range 0 to 9999
                    _lastWrittenPos = (int)Mathf.MoveTowards(_lastWrittenPos, targetPos, 400);
                }
                else
                {
                    _lastWrittenPos = targetPos;
                    _lastWrittenDelta = int.Parse(prt.Substring(7));
                    if (_bigDeltaDetected)
                    {
                        _bigDeltaDetected = false;
                        //SuperController.LogMessage("BIG_DELTA_RESTORED");
                    }
                }

                if (_loadingStatus == LOAD_STATUS_WAIT_ONE_WRITE)
                {
                    _loadingStatus = LOAD_STATUS_WAIT_ONE_UPDATE;
                    // SuperController.LogMessage("LOAD_STATUS_WAIT_ONE_UPDATE");
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("parse error " + e.ToString());
                return;
            }
        }
        public void Update(bool tracking)
        {
            UpdateBackroll(tracking);

            PrepareOptions();
            UpdateMotionSource();
            SetOptions();
            
            UpdateNetwork();
            UpdateSimulator();
        }

        public void OnSceneChanging() 
        {
            _loadingStatus = LOAD_STATUS_SCENE_LOADING;
            // SuperController.LogMessage("LOAD_STATUS_SCENE_LOADING");
        }
        public void OnSceneChanged() 
        {
            _loadingStatus = LOAD_STATUS_WAIT_ONE_WRITE;
            // SuperController.LogMessage("LOAD_STATUS_WAIT_ONE_WRITE");
        }

        // back the launch device if no tracking has been registered for about 2 seconds
        private void UpdateBackroll(bool tracking)
        {
            _tracking = tracking;
            _resetting = false;
            if (!_tracking || _loadingStatus != LOAD_STATUS_NONE)
            {
                if ((Time.time - _udpatesSinceLastTracking) > 2f || _loadingStatus != LOAD_STATUS_NONE)
                {
                    _resetting = true;
                    _lastWrittenPos = Math.Max(0, _lastWrittenPos - 60);
                    _slowInitiate = Time.time;
                }
            } 
            else
            {
                _udpatesSinceLastTracking = Time.time;
            }
        }

        private void UpdateNetwork()
        {
            if (_network == null)
                return;

            _networkPollTimer -= Time.deltaTime;
            if (_networkPollTimer <= 0.0f)
            {
                ReceiveNetworkMessages();
                _networkPollTimer = NETWORK_LISTEN_INTERVAL - Mathf.Min(-_networkPollTimer, NETWORK_LISTEN_INTERVAL);
            }
        }

        // In case of future two-way communication
        private void ReceiveNetworkMessages()
        {
            byte[] msg = _network.GetNextMessage();
            if (msg != null && msg.Length > 0)
            {
                //SuperController.LogMessage(msg[0].ToString());
            }
        }

        private void UpdateSimulator()
        {
            var prevPos = _simulatorPosition.val;

            var newPos = Mathf.MoveTowards(prevPos, _simulatorTarget,
                LaunchUtils.PredictDistanceTraveled(_simulatorSpeed, Time.deltaTime));

            _simulatorPosition.SetVal(newPos);

            if (_currentMotionSource != null)
            {
                _currentMotionSource.OnSimulatorUpdate(prevPos, newPos, Time.deltaTime);
            }
        }

        private void SetSimulatorTarget(float pos, float speed)
        {
            _simulatorTarget = Mathf.Clamp(pos, 0.0f, LaunchUtils.LAUNCH_MAX_VAL);
            _simulatorSpeed = Mathf.Clamp(speed, 0.0f, LaunchUtils.LAUNCH_MAX_VAL);
        }

        private void UpdateMotionSource()
        {
            if (_currentMotionSource != null)
            {
                byte pos = 0;
                byte speed = 0;

                float simulatedPos = (((float)_lastWrittenPos) / 9999f);
                if (_invertToggle.val)
                {
                    simulatedPos = 1f - simulatedPos;
                }
                if (_currentMotionSource.OnUpdate(simulatedPos, ref pos, ref speed))
                {
                    if (_loadingStatus == LOAD_STATUS_WAIT_ONE_UPDATE)
                    {
                        _loadingStatus = LOAD_STATUS_NONE;
                        // SuperController.LogMessage("LOAD_STATUS_NONE");
                    }
                    if ((Time.time - _slowInitiate) < 2f || _bigDeltaDetected || _loadingStatus != LOAD_STATUS_NONE)
                    {
                        // (Time.time - _udpatesSinceLastTracking) < 2f
                        // cap speed when tracking from a pause or big delta, to avoid the device going from 0 to max because of a single touch in the animation
                        speed = (byte)Mathf.Min(speed, LaunchUtils.LAUNCH_MAX_SPEED / 4f);
                        // SuperController.LogMessage("Slow Speed Cap");
                    }
                    if (_resetting && speed > 20)
                    {
                        speed = 20;
                    }
                    if (speed > _launchMaxSpeed.val)
                    {
                        speed = (byte)_launchMaxSpeed.val;
                    }
                    SendLaunchPosition(pos, speed);
                }
            }
        }

        private static byte[] _launchData = new byte[6];
        private void SendLaunchPosition(byte pos, byte speed)
        {
            SetSimulatorTarget(pos, speed);

            if (_network == null)
            {
                return;
            }

            //if (!_pauseLaunchMessages.val)
            //{
            _launchData[0] = pos;
            _launchData[1] = speed;

            float dist = Mathf.Abs(pos - _lastSentLaunchPos);
            float duration = LaunchUtils.PredictMoveDuration(dist, speed);

            var durationData = BitConverter.GetBytes(duration);
            _launchData[2] = durationData[0];
            _launchData[3] = durationData[1];
            _launchData[4] = durationData[2];
            _launchData[5] = durationData[3];

            //SuperController.LogMessage(string.Format("Sending: P:{0}, S:{1}, D:{2}", pos, speed, duration));

            _network.Send(_launchData, _launchData.Length);

            _lastSentLaunchPos = pos;
            //}
        }

        protected virtual void Dispose(bool disposing)
        {
            StopUdp();
        }

        

        //private byte _lastSentLaunchPos;
        //private int _lastDirectionSign;
        //private float _accumulatedTime = 0f;

        //private int _speedAvg5Index = 0;
        //private float[] _speedAvg5 = new float[3];
        
        //private byte[] GetLaunchData(byte pos, float deltaTimeMs)
        //{

        //    _accumulatedTime += deltaTimeMs;
            

        //    if (pos < 0)
        //        pos = 0;
        //    if (pos > 100)
        //        pos = 100;

        //    byte resultingPos = pos;

        //    float posDelta = pos - _lastSentLaunchPos;

        //    // float posDelta = pos - _lastSentLaunchPos;
        //    // predict position one step ahead
        //    //int posPrediction = ((int)Math.Round(posDelta));
        //    //if (posPrediction > 100)
        //    //    posPrediction = 100;
        //    //if (posPrediction < -100)
        //    //    posPrediction = -100;

        //    //if ((pos + posPrediction) < 0)
        //    //    resultingPos = 0;
        //    //else if ((pos + posPrediction) > 100)
        //    //    resultingPos = 100;
        //    //else
        //    //    resultingPos = (byte)(pos + (byte)posPrediction);

        //    int sign = Math.Sign(posDelta);
        //    posDelta = Math.Abs(posDelta);

        //    bool forceUpdate = _accumulatedTime >= 200f || posDelta > 5 && _lastDirectionSign != sign && _accumulatedTime >= 20f;
        //    if (!forceUpdate)
        //        return null;

        //    _lastDirectionSign = sign;

        //    float secondsSinceLastUpdate = _accumulatedTime / 900f;

        //    float speed = LaunchUtils.PredictMoveSpeed(resultingPos, resultingPos + posDelta, secondsSinceLastUpdate); // * _lagFactor.Factor;

        //    _speedAvg5[_speedAvg5Index] = speed;
        //    _speedAvg5Index++;
        //    if (_speedAvg5Index >= _speedAvg5.Length)
        //        _speedAvg5Index = 0;

        //    speed = 0f;
        //    for (int i = 0; i < _speedAvg5.Length; i++)
        //        speed += _speedAvg5[i];
        //    speed /= _speedAvg5.Length;

        //    //if (posDelta > 5)
        //    //{
        //    //    resultingPos = (byte)Math.Min(100, Math.Max(0, ((int)resultingPos + Math.Round(posDelta * (float)sign))));
        //    //}
        //    //float deltaToNext = 0f;
        //    //if (speed > 12f && posDelta > 5) // LaunchUtils.PredictMoveSpeed clamps to 10 and 90, which makes predicting the next based on speed on slow movements not possible using the "speed" variable.
        //    //{
        //    //    deltaToNext = Math.Max(0, LaunchUtils.PredictDistanceTraveled(speed, secondsSinceLastUpdate));
        //    //    // predict next position based on current speed
        //    //    resultingPos = (byte)Math.Min(100, Math.Max(0, ((int)resultingPos + Math.Round(deltaToNext * (float)sign))));
        //    //}

        //    // predict slightly ahead based on current sign
        //    resultingPos = (byte)Math.Min(100, Math.Max(0, ((int)resultingPos + Math.Round(10f * (float)sign))));

        //    float duration = secondsSinceLastUpdate;

        //    byte[] launchData = new byte[6];
        //    launchData[0] = resultingPos;
        //    launchData[1] = (byte)Math.Round(speed);

        //    var durationData = BitConverter.GetBytes(duration);
        //    launchData[2] = durationData[0];
        //    launchData[3] = durationData[1];
        //    launchData[4] = durationData[2];
        //    launchData[5] = durationData[3];

        //    _lastSentLaunchPos = pos;
        //    _accumulatedTime = 0f;

        //    SuperController.LogMessage(string.Format("Sending: P:{0}, {1}, S:{2}, D:{3}", pos, resultingPos, speed, duration));
        //    return launchData;
        //}

        public void Dispose()
        {
            Dispose(disposing: true);
            GC.SuppressFinalize(this);
        }
    }
}
